#include ".\level.h"
#include ".\LD12.h"

#include <debug/debugclient.h>


Level::Level()
{

  Clear();

}

Level::~Level(void)
{
}



void Level::Render()
{

  theApp.m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );

  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_SELECTARG1 );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );

  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1 );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );

  int   Offset = (int)m_LevelDisplayOffset;

  int   StartY = Offset / 32;
  int   DY = Offset % 32;

  for ( int i = 0; i < 20; ++i )
  {
    for ( int j = 0; j < 16; ++j )
    {
      theApp.RenderTextureSectionAsTile( i * 32, j * 32 - DY, theApp.m_Tiles[GetField( i, StartY + j )] );
    }
  }

  theApp.m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );

  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );

  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_MODULATE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );

  // Render Objects
  tListObjects::iterator it = m_Objects.begin();
  while ( it != m_Objects.end() )
  {
    GameObject& Obj( *it );

    if ( !Obj.m_Hidden )
    {
      Obj.Render( *this );
    }

    ++it;
  }

  // Render FX
  theApp.m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
  theApp.m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_ONE );
  theApp.m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );

  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );

  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_MODULATE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );

  // Render Objects
  it = m_Objects.begin();
  while ( it != m_Objects.end() )
  {
    GameObject& Obj( *it );

    if ( !Obj.m_Hidden )
    {
      Obj.RenderFX( *this );
    }

    ++it;
  }

}



void Level::DamageInnerStructure( int X, int Y, float Damage )
{

  LD12::TileTypes   Tile = GetField( X, Y );

  if ( ( Tile != LD12::TILE_EMPTY )
  &&   ( Tile != LD12::TILE_SKY_EMPTY )
  &&   ( Tile != LD12::TILE_GRASS ) )
  {
    if ( Damage >= Structure( X, Y ).InnerStrength )
    {
      // auto-break strength below
      Structure( X, Y ).InnerStrength = 0;
    }
    else
    {
      Structure( X, Y ).InnerStrength -= (int)Damage;
    }
  }

}



void Level::ShatterGlass( int X, int Y )
{

  int   ParticleCount = 1 + rand() % 4;
  for ( int i = 0; i < ParticleCount; ++i )
  {
    GameObject* pParticle = CreateObject( LD12::GO_PARTICLE_GLASS, X * 32 + rand() % 32, Y * 32 + rand() % 32 );
    pParticle->m_AccX = 31.0f - rand() % 63;
    pParticle->m_AccY = 5.0f - rand() % 31;
    pParticle->m_Angle = (float)( rand() % 360 );
    pParticle->m_AngleDelta = 90.0f - rand() % 181;
    CollateralParticle( pParticle );
  }

}



void Level::ShatterBeton( int X, int Y )
{

  int   ParticleCount = 1 + rand() % 4;
  for ( int i = 0; i < ParticleCount; ++i )
  {
    GameObject* pParticle = CreateObject( LD12::GO_PARTICLE_BETON, X * 32 + rand() % 32, Y * 32 + rand() % 32 );
    pParticle->m_AccX = 15.0f - rand() % 31;
    pParticle->m_AccY = 5.0f - rand() % 31;
    pParticle->m_Angle = (float)( rand() % 360 );
    pParticle->m_AngleDelta = 45.0f - rand() % 91;
    CollateralParticle( pParticle );
  }

}



void Level::CollateralParticle( GameObject* pObj )
{

  // check chance for collateral particle
  if ( ( rand() % 100 ) >= 30 )
  {
    pObj->m_Collateral = true;
    pObj->m_AccX *= 5.0f;
    if ( pObj->m_Type == LD12::GO_PARTICLE_BETON )
    {
      pObj->m_AccX *= 3.0f;
      if ( ( rand() % 100 ) >= 98 )
      {
        if ( rand() % 2 )
        {
          pObj->m_Texture = LD12::TEX_SINK;
        }
        else
        {
          pObj->m_Texture = LD12::TEX_TOILET;
        }
        pObj->m_LifeTime *= 3.0f;
      }
    }
    // let the particle fly a bit before smashing something
    pObj->m_ParticleHurtDelay = 0.25f + ( rand() % 5 ) * 0.1f;
  }

}



void Level::DamageStructure( int X, int Y, float Damage )
{

  LD12::TileTypes   Tile = GetField( X, Y );

  if ( ( Tile != LD12::TILE_EMPTY )
  &&   ( Tile != LD12::TILE_SKY_EMPTY )
  &&   ( Tile != LD12::TILE_GRASS ) )
  {
    if ( Damage >= Structure( X, Y ).Strength )
    {
      // auto-break strength below
      Structure( X, Y ).Strength = 0;
      if ( ( Tile == LD12::TILE_WINDOW_1 )
      ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_1 ) )
      {
        SetField( X, Y, (LD12::TileTypes)( LD12::TILE_WINDOW_1_BROKEN_2 + rand() % 4 ) );

        m_SoundRequests.insert( rand() % 2 ? LD12::SOUND_BREAK_GLASS_1 : LD12::SOUND_BREAK_GLASS_2 );
      }
      else if ( ( Tile == LD12::TILE_OTHER_WINDOW_1 )
      ||        ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_1 ) )
      {
        SetField( X, Y, (LD12::TileTypes)( LD12::TILE_OTHER_WINDOW_1_BROKEN_2 + rand() % 4 ) );

        m_SoundRequests.insert( rand() % 2 ? LD12::SOUND_BREAK_GLASS_1 : LD12::SOUND_BREAK_GLASS_2 );
      }
      else if ( Tile == LD12::TILE_PLASTIC_FENCE )
      {
        // rip plastic
        SetField( X, Y, LD12::TILE_PLASTIC_FENCE_RIPPED );

        m_SoundRequests.insert( rand() % 2 ? LD12::SOUND_HIT_PLASTIC_FENCE_1 : LD12::SOUND_HIT_PLASTIC_FENCE_2 );
      }
      else if ( Tile == LD12::TILE_WOODEN_FENCE )
      {
        SetField( X, Y, LD12::TILE_WOODEN_FENCE_BROKEN );

        m_SoundRequests.insert( rand() % 2 ? LD12::SOUND_HIT_WOODEN_FENCE_1 : LD12::SOUND_HIT_WOODEN_FENCE_2 );
      }
      else if ( Tile == LD12::TILE_WOODEN_FENCE_TOP )
      {
        SetField( X, Y, LD12::TILE_WOODEN_FENCE_TOP_BROKEN );

        m_SoundRequests.insert( rand() % 2 ? LD12::SOUND_HIT_WOODEN_FENCE_1 : LD12::SOUND_HIT_WOODEN_FENCE_2 );
      }
    }
    else
    {
      Structure( X, Y ).Strength -= (int)Damage;
      if ( Tile == LD12::TILE_WINDOW_1 )
      {
        if ( Structure( X, Y ).Strength <= 50 )
        {
          if ( Structure( X, Y ).Strength + (int)Damage > 50 )
          {
            // glass shattered now
            ShatterGlass( X, Y );
            m_SoundRequests.insert( rand() % 2 ? LD12::SOUND_BREAK_GLASS_1 : LD12::SOUND_BREAK_GLASS_2 );
          }
          SetField( X, Y, LD12::TILE_WINDOW_1_BROKEN_1 );
        }
      }
      else if ( Tile == LD12::TILE_OTHER_WINDOW_1 )
      {
        if ( Structure( X, Y ).Strength <= 50 )
        {
          if ( Structure( X, Y ).Strength + (int)Damage > 50 )
          {
            // glass shattered now
            ShatterGlass( X, Y );
            m_SoundRequests.insert( rand() % 2 ? LD12::SOUND_BREAK_GLASS_1 : LD12::SOUND_BREAK_GLASS_2 );
          }
          SetField( X, Y, LD12::TILE_OTHER_WINDOW_1_BROKEN_1 );
        }
      }

    }
  }

}



void Level::CollapseLevel()
{

  for ( int i = 0; i < 20; ++i )
  {
    for ( int j = 0; j < m_Height; ++j )
    {
      LD12::TileTypes Tile = GetField( i, j );
      if ( ( Tile != LD12::TILE_EMPTY )
      &&   ( Tile != LD12::TILE_SKY_EMPTY ) )
      {
        TileInfo  Struct = Structure( i, j );
        if ( Struct.InnerStrength == 0 )
        {
          // calculate chance of breakdown
          if ( ( rand() % 100 ) >= 95 )
          {
            ShatterBeton( i, j );
            m_SoundRequests.insert( (LD12::SoundTypes)( LD12::SOUND_COLLAPSE_1 + rand() % 3 ) );
            if ( j > 0 )
            {
              if ( ( Tile == LD12::TILE_WINDOW_1 )
              ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_1 )
              ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_2 )
              ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_3 )
              ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_4 )
              ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_5 ) )
              {
                --m_CurrentHouseTiles;
              }
              if ( ( Tile == LD12::TILE_OTHER_WINDOW_1 )
              ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_1 )
              ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_2 )
              ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_3 )
              ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_4 )
              ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_5 ) )
              {
                --m_CurrentOtherHouseTiles;
              }
              for ( int k = j - 1; k >= 0; --k )
              {
                TileInfo          StructAbove = Structure( i, k );
                LD12::TileTypes   TileAbove = GetField( i, k );

                SetField( i, k + 1, TileAbove );
                Structure( i, k + 1 ) = StructAbove;
                DamageStructure( i, k + 1, 5.0f + ( rand() % 15 ) );

                Level::tListObjects::iterator   itObj( m_Objects.begin() );
                while ( itObj != m_Objects.end() )
                {
                  GameObject&   Obj( *itObj );

                  if ( ( Obj.m_Type == LD12::GO_TIMEBOMB )
                  &&   ( Obj.m_DispX / 32 == i )
                  &&   ( Obj.m_DispY / 32 == k ) )
                  {
                    // bombs fall with the building
                    Obj.m_DispY += 32;
                  }
                  ++itObj;
                }
              }
              // damage tile below depending on the tiles above
              if ( j + 1 < m_Height )
              {
                LD12::TileTypes   TileBelow = GetField( i, j + 1 );

                if ( ( TileBelow != LD12::TILE_EMPTY )
                &&   ( TileBelow != LD12::TILE_SKY_EMPTY )
                &&   ( TileBelow != LD12::TILE_GRASS ) )
                {
                  int     TileCountAbove = 0;

                  for ( int k = 0; k <= j; ++k )
                  {
                    LD12::TileTypes   TileAbove = GetField( i, k );
                    if ( ( TileAbove != LD12::TILE_EMPTY )
                    &&   ( TileAbove != LD12::TILE_SKY_EMPTY )
                    &&   ( TileAbove != LD12::TILE_GRASS ) )
                    {
                      TileCountAbove++;
                    }
                  }
                  if ( TileCountAbove > 0 )
                  {
                    // calculate random damage
                    DamageInnerStructure( i, j + 1, 45.0f * TileCountAbove );
                    CreateObject( LD12::GO_SMOKE_LEFT, i * 32 - 60 + 32, j * 32 - 60 + 32 + 32 );
                    CreateObject( LD12::GO_SMOKE_RIGHT, i * 32 + 31 + 32, j * 32 - 60 + 32 + 32 );
                  }
                }
              }
            }
            SetField( i, 0, LD12::TILE_SKY_EMPTY );
            Structure( i, 0 ) = TileInfo();
          }
        }
      }
    }
  }

}



void Level::Update( float ElapsedTime )
{

  m_PhysicsStep += ElapsedTime;

  while ( m_PhysicsStep >= 0.02f )
  {
    tListObjects::iterator    it( m_Objects.begin() );
    while ( it != m_Objects.end() )
    {
      GameObject& Obj( *it );

      Obj.UpdatePhysicsStep( *this );
      if ( ( Obj.m_DispX < -60 )
      ||   ( Obj.m_DispY < -60 )
      ||   ( Obj.m_DispX > 640 + 60 )
      ||   ( Obj.m_DispY > 60 + m_Height * 32 ) )
      {
        // kill object if outside the screen so we don't have to wait too long
        Obj.m_RemoveMe = true;
      }
      if ( Obj.m_RemoveMe )
      {
        it = m_Objects.erase( it );
      }
      else
      {
        ++it;
      }
    }
    CollapseLevel();

    m_PhysicsStep -= 0.02f;
  }

  tListObjects::iterator    it( m_Objects.begin() );
  while ( it != m_Objects.end() )
  {
    GameObject& Obj( *it );

    Obj.Update( ElapsedTime, *this );
    if ( Obj.m_RemoveMe )
    {
      it = m_Objects.erase( it );
    }
    else
    {
      ++it;
    }
  }

  it = m_Objects.begin();
  while ( it != m_Objects.end() )
  {
    GameObject&   Obj( *it );

    ++it;
  }

  PlayRequestedSounds();

}



void Level::PlayRequestedSounds()
{

  std::set<LD12::SoundTypes>::iterator    itSound( m_SoundRequests.begin() );
  while ( itSound != m_SoundRequests.end() )
  {
    theApp.PlaySound( *itSound );

    ++itSound;
  }
  m_SoundRequests.clear();

}



void Level::Clear()
{

  m_LevelDisplayOffset  = 0.0f;
  m_PhysicsStep         = 0.0f;
  m_MoneyAvailable      = 0;
  m_Objects.clear();
  m_SoundRequests.clear();
  m_InitialHouseTiles   = 0;
  m_CurrentHouseTiles   = 0;
  m_InitialOtherHouseTiles  = 0;
  m_CurrentOtherHouseTiles  = 0;
  m_PercentageNeeded    = 25;

  m_Height              = 14;

  m_Field.resize( 20 * m_Height );
  m_Structure.resize( 20 * m_Height );
  for ( size_t i = 0; i < m_Field.size(); ++i )
  {
    m_Field[i]      = LD12::TILE_SKY_EMPTY;
    m_Structure[i]  = TileInfo();
  }
  m_Name.erase();
  m_ObjIndex = 1;

}



bool Level::Load( int Nr )
{

  Clear();

  char    Temp[260];

  wsprintf( Temp, "data\\level%d.dat", Nr );

  FILE*   In = fopen( Temp, "rb" );
  if ( In == NULL )
  {
    return false;
  }

  int   Width = 0;
  fread( &Width, sizeof( Width ), 1, In );
  fread( &m_Height, sizeof( m_Height ), 1, In );
  m_Field.resize( Width * m_Height );
  m_Structure.resize( Width * m_Height );
  fread( &m_Field[0], sizeof( LD12::TileTypes ) * m_Field.size(), 1, In );

  for ( size_t i = 0; i < m_Structure.size(); ++i )
  {
    m_Structure[i]  = TileInfo();
  }

  int   Count = 0;
  fread( &Count, sizeof( Count ), 1, In );
  for ( int i = 0; i < Count; ++i )
  {
    LD12::GameObjectType    ObjType;

    fread( &ObjType, sizeof( ObjType ), 1, In );

    GameObject*    pObjNew = CreateObject( ObjType );

    fread( &pObjNew->m_DispX, sizeof( pObjNew->m_DispX ), 1, In );
    fread( &pObjNew->m_DispY, sizeof( pObjNew->m_DispY ), 1, In );
    fread( &pObjNew->m_LifeTime, sizeof( pObjNew->m_LifeTime ), 1, In );
    fread( &pObjNew->m_Number, sizeof( pObjNew->m_Number ), 1, In );
  }
  fread( &m_MoneyAvailable, sizeof( m_MoneyAvailable ), 1, In );
  int   Size = 0;
  fread( &Size, sizeof( Size ), 1, In );
  if ( Size )
  {
    char*    pBuffer = new char[Size];
    fread( pBuffer, Size, 1, In );
    m_Name.append( pBuffer, Size );
    delete[] pBuffer;
  }

  fread( &m_PercentageNeeded, sizeof( m_PercentageNeeded ), 1, In );
  fclose( In );

  m_InitialHouseTiles = 0;
  m_InitialOtherHouseTiles = 0;
  for ( size_t i = 0; i < m_Field.size(); ++i )
  {
    LD12::TileTypes Tile( m_Field[i] );

    if ( ( Tile == LD12::TILE_WINDOW_1 )
    ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_1 )
    ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_2 )
    ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_3 )
    ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_4 )
    ||   ( Tile == LD12::TILE_WINDOW_1_BROKEN_5 ) )
    {
      ++m_InitialHouseTiles;
    }
    if ( ( Tile == LD12::TILE_OTHER_WINDOW_1 )
    ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_1 )
    ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_2 )
    ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_3 )
    ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_4 )
    ||   ( Tile == LD12::TILE_OTHER_WINDOW_1_BROKEN_5 ) )
    {
      ++m_InitialOtherHouseTiles;
    }
  }
  m_CurrentHouseTiles = m_InitialHouseTiles;
  m_CurrentOtherHouseTiles = m_InitialOtherHouseTiles;

  return true;

}



bool Level::Save( int Nr )
{

  char    Temp[260];

  wsprintf( Temp, "data\\level%d.dat", Nr );

  FILE*   Out = fopen( Temp, "wb" );
  if ( Out == NULL )
  {
    return false;
  }

  int   Width = 20;
  fwrite( &Width, sizeof( Width ), 1, Out );
  fwrite( &m_Height, sizeof( m_Height ), 1, Out );
  fwrite( &m_Field[0], sizeof( LD12::TileTypes ) * m_Field.size(), 1, Out );

  int   Count = (int)m_Objects.size();
  fwrite( &Count, sizeof( Count ), 1, Out );

  tListObjects::iterator    itObj( m_Objects.begin() );
  while ( itObj != m_Objects.end() )
  {
    GameObject& Obj( *itObj );

    fwrite( &Obj.m_Type, sizeof( Obj.m_Type ), 1, Out );

    fwrite( &Obj.m_DispX, sizeof( Obj.m_DispX ), 1, Out );
    fwrite( &Obj.m_DispY, sizeof( Obj.m_DispY ), 1, Out );
    fwrite( &Obj.m_LifeTime, sizeof( Obj.m_LifeTime ), 1, Out );
    fwrite( &Obj.m_Number, sizeof( Obj.m_Number ), 1, Out );

    ++itObj;
  }

  fwrite( &m_MoneyAvailable, sizeof( m_MoneyAvailable ), 1, Out );
  int   Size = (int)m_Name.length();
  fwrite( &Size, sizeof( Size ), 1, Out );
  fwrite( m_Name.c_str(), Size, 1, Out );
  fwrite( &m_PercentageNeeded, sizeof( m_PercentageNeeded ), 1, Out );
  fclose( Out );

  return true;

}



GameObject* Level::CreateObject( LD12::GameObjectType ObjType, int X, int Y )
{

  GameObject    Obj;

  Obj.m_Type = ObjType;

  switch ( Obj.m_Type )
  {
    case LD12::GO_TIMEBOMB:
      Obj.m_Texture = LD12::TEX_BOMB;
      Obj.m_Power   = 150;
      break;
    case LD12::GO_PLASTIC_FENCE:
      Obj.m_Texture = LD12::TEX_ITEM_PLASTIC_FENCE;
      Obj.m_Power   = 25;
      break;
    case LD12::GO_WOODEN_FENCE:
      Obj.m_Texture = LD12::TEX_ITEM_WOODEN_FENCE;
      Obj.m_Power   = 75;
      break;
    case LD12::GO_SMOKE_LEFT:
      Obj.m_Texture = LD12::TEX_SMOKE_LEFT;
      Obj.m_LifeTime = 1.0f + ( rand() % 10 ) * 0.1f;
      break;
    case LD12::GO_SMOKE_RIGHT:
      Obj.m_Texture = LD12::TEX_SMOKE_RIGHT;
      Obj.m_LifeTime = 1.0f + ( rand() % 10 ) * 0.1f;
      break;
    case LD12::GO_PARTICLE_GLASS:
      Obj.m_Texture = LD12::TEX_PARTICLE_GLASS_1;
      Obj.m_LifeTime = 3.0f + ( rand() % 10 ) * 0.1f;
      Obj.m_Power = 15;
      break;
    case LD12::GO_PARTICLE_BETON:
      if ( rand() % 2 )
      {
        Obj.m_Texture = LD12::TEX_PARTICLE_BETON_1;
      }
      else
      {
        Obj.m_Texture = LD12::TEX_PARTICLE_BETON_2;
      }
      Obj.m_LifeTime = 5.0f + ( rand() % 10 ) * 0.1f;
      Obj.m_Power = 50;
      break;
  }

  Obj.m_ObjIndex = m_ObjIndex;
  ++m_ObjIndex;

  if ( ( X != -100 )
  &&   ( Y != -100 ) )
  {
    Obj.m_DispX = X;
    Obj.m_DispY = Y;
  }

  m_Objects.push_back( Obj );

  return &m_Objects.back();

}



void Level::RemoveObject( GameObject* pObj )
{

  tListObjects::iterator    itObj( m_Objects.begin() );
  while ( itObj != m_Objects.end() )
  {
    GameObject& Obj( *itObj );

    if ( &Obj == pObj )
    {
      m_Objects.erase( itObj );
      return;
    }

    ++itObj;
  }

}



bool Level::IsBlocked( GameObject* pObj, int X, int Y, int W, int H, int& BlockedX, int& BlockedY )
{

  X -= W / 2;
  Y -= H / 2;

  int     X1 = X / 32;
  int     X2 = ( X + W - 1 ) / 32;
  int     Y1 = Y / 32;
  int     Y2 = ( Y + H - 1 ) / 32;

  bool    Blocked = false;

  for ( int x = X1; x <= X2; ++x )
  {
    for ( int y = Y1; y <= Y2; ++y )
    {
      LD12::TileTypes Tile = GetField( x, y );

      if ( ( Tile != LD12::TILE_EMPTY )
      &&   ( Tile != LD12::TILE_SKY_EMPTY )
      &&   ( Tile != LD12::TILE_PLASTIC_FENCE_RIPPED )
      &&   ( Tile != LD12::TILE_WOODEN_FENCE_BROKEN )
      &&   ( Tile != LD12::TILE_WOODEN_FENCE_TOP_BROKEN ) )
      {
        BlockedX = x;
        BlockedY = y;
        return true;
      }
    }
  }

  return Blocked;

}



bool Level::CanAreaContainBomb( int X, int Y, int W, int H )
{

  X -= W / 2;
  Y -= H / 2;

  int     X1 = X / 20;
  int     X2 = ( X + W - 1 ) / 20;
  int     Y1 = Y / 20;
  int     Y2 = ( Y + H - 1 ) / 20;

  for ( int x = X1; x <= X2; ++x )
  {
    for ( int y = Y1; y <= Y2; ++y )
    {
      if ( ( x < 0 )
      ||   ( x >= 32 )
      ||   ( y < 0 )
      ||   ( y >= 22 ) )
      {
        return false;
      }
      else
      {
        LD12::TileTypes Tile = GetField( x, y );

        if ( Tile != LD12::TILE_EMPTY )
        {
          return false;
        }
      }
    }
  }

  return true;

}



bool Level::CanObjectBeBlocked( GameObject* pObj )
{

  return ( ( pObj->m_Collateral )
  &&       ( pObj->m_ParticleHurtDelay == 0.0f ) );

}



void Level::OnObjectBlocked( GameObject* pObj, int TX, int TY )
{

  if ( ( pObj->m_Type == LD12::GO_PARTICLE_BETON )
  ||   ( pObj->m_Type == LD12::GO_PARTICLE_GLASS ) )
  {
    LD12::TileTypes   Tile = GetField( TX, TY );
    TileInfo&         Struct = Structure( TX, TY );

    if ( Struct.Strength == 0 )
    {
      DamageInnerStructure( TX, TY, (float)pObj->m_Power );
    }
    else
    {
      DamageStructure( TX, TY, (float)pObj->m_Power );
    }
  }
  pObj->m_RemoveMe = true;

}



bool Level::MoveObject( GameObject* pObj, float DX, float DY )
{

  pObj->m_FractX += DX;
  pObj->m_FractY += DY;

  if ( !CanObjectBeBlocked( pObj ) )
  {
    while ( pObj->m_FractX >= 1.0f )
    {
      pObj->m_DispX++;
      pObj->m_FractX -= 1.0f;
    }
    while ( pObj->m_FractX <= -1.0f )
    {
      pObj->m_DispX--;
      pObj->m_FractX += 1.0f;
    }
    while ( pObj->m_FractY >= 1.0f )
    {
      pObj->m_DispY++;
      pObj->m_FractY -= 1.0f;
    }
    while ( pObj->m_FractY <= -1.0f )
    {
      pObj->m_DispY--;
      pObj->m_FractY += 1.0f;
    }
    return true;
  }

  bool Blocked = false;

  int   BlockedX = -1;
  int   BlockedY = -1;

  while ( ( pObj->m_FractX >= 1.0f )
  ||      ( pObj->m_FractX <= -1.0f )
  ||      ( pObj->m_FractY <= -1.0f )
  ||      ( pObj->m_FractY >= 1.0f ) )
  {
    if ( pObj->m_FractX <= -1.0f )
    {
      // go left
      if ( IsBlocked( pObj, (int)pObj->m_DispX - 1, pObj->m_DispY, 25, 25, BlockedX, BlockedY ) )
      {
        // blocked left
        pObj->m_FractX = 0.0f;
        pObj->m_SpeedX = -pObj->m_SpeedX;
        Blocked = true;
        OnObjectBlocked( pObj, BlockedX, BlockedY );
        if ( pObj->m_RemoveMe )
        {
          return false;
        }
      }
      else
      {
        pObj->m_DispX--;
        pObj->m_FractX += 1.0f;
      }
    }
    else if ( pObj->m_FractX >= 1.0f )
    {
      // go right
      if ( IsBlocked( pObj, (int)pObj->m_DispX + 1, pObj->m_DispY, 25, 25, BlockedX, BlockedY ) )
      {
        // blocked right
        pObj->m_FractX = 0.0f;
        pObj->m_SpeedX = -pObj->m_SpeedX;
        Blocked = true;
        OnObjectBlocked( pObj, BlockedX, BlockedY );
        if ( pObj->m_RemoveMe )
        {
          return false;
        }
      }
      else
      {
        pObj->m_DispX++;
        pObj->m_FractX -= 1.0f;
      }
    }
    if ( pObj->m_FractY <= -1.0f )
    {
      // go up
      if ( IsBlocked( pObj, pObj->m_DispX, pObj->m_DispY - 1, 25, 25, BlockedX, BlockedY ) )
      {
        // blocked up
        pObj->m_FractY = 0.0f;
        pObj->m_SpeedY = -pObj->m_SpeedY;
        Blocked = true;
        OnObjectBlocked( pObj, BlockedX, BlockedY );
        if ( pObj->m_RemoveMe )
        {
          return false;
        }
      }
      else
      {
        pObj->m_DispY--;
        pObj->m_FractY += 1.0f;
      }
    }
    else if ( pObj->m_FractY >= 1.0f )
    {
      // go down
      if ( IsBlocked( pObj, pObj->m_DispX, pObj->m_DispY + 1, 25, 25, BlockedX, BlockedY ) )
      {
        // blocked down
        pObj->m_FractY = 0.0f;
        pObj->m_SpeedY = -pObj->m_SpeedY;
        Blocked = true;
        OnObjectBlocked( pObj, BlockedX, BlockedY );
        if ( pObj->m_RemoveMe )
        {
          return false;
        }
      }
      else
      {
        pObj->m_DispY++;
        pObj->m_FractY -= 1.0f;
      }
    }
  }
  return !Blocked;

}



void Level::SetField( int X, int Y, LD12::TileTypes Tile )
{

  if ( ( X < 0 )
  ||   ( Y < 0 ) )
  {
    return;
  }
  if ( X + 20 * Y >= (int)m_Field.size() )
  {
    return;
  }
  m_Field[X + 20 * Y] = Tile;

}



LD12::TileTypes Level::GetField( int X, int Y )
{

  if ( ( X < 0 )
  ||   ( Y < 0 ) )
  {
    return LD12::TILE_EMPTY;
  }
  if ( X + 20 * Y >= (int)m_Field.size() )
  {
    return LD12::TILE_EMPTY;
  }
  return m_Field[X + 20 * Y];

}



Level::TileInfo& Level::Structure( int X, int Y )
{

  if ( X + 20 * Y >= (int)m_Field.size() )
  {
    static TileInfo   Dummy;
    return Dummy;
  }
  return m_Structure[X + 20 * Y];

}



void Level::Resize( int NewHeight )
{

  m_Field.resize( NewHeight * 20 );
  m_Structure.resize( NewHeight * 20 );
  m_Height = NewHeight;

}



void Level::Explode( GameObject* pBomb )
{

  // TODO - Damage structure
  float     BombRadius    = 80.0f;
  float     BombStrength  = pBomb->m_Power * ( 8.0f + ( rand() % 4 ) ) * 0.1f;

  // Circular Bomb damage radius
  int   X1 = ( pBomb->m_DispX - (int)BombRadius ) / 32;
  int   Y1 = ( pBomb->m_DispY - (int)BombRadius ) / 32;
  int   X2 = ( pBomb->m_DispX + (int)BombRadius ) / 32;
  int   Y2 = ( pBomb->m_DispY + (int)BombRadius ) / 32;

  for ( int X = X1; X <= X2; ++X )
  {
    for ( int Y = Y1; Y <= Y2; ++Y )
    {
      LD12::TileTypes   Tile = GetField( X, Y );

      // can there be damage at all?
      if ( ( Tile != LD12::TILE_EMPTY )
      &&   ( Tile != LD12::TILE_SKY_EMPTY )
      &&   ( Tile != LD12::TILE_GRASS ) )
      {
        float   TileCenterX = X * 32.0f + 15.0f;
        float   TileCenterY = Y * 32.0f + 15.0f;

        float   TrueDistance = sqrtf( ( pBomb->m_DispX - TileCenterX ) * ( pBomb->m_DispX - TileCenterX ) + ( pBomb->m_DispY - TileCenterY ) * ( pBomb->m_DispY - TileCenterY ) );

        if ( TrueDistance <= BombRadius )
        {
          // close enough to actually take damage
          float CurBombStrength = BombStrength * ( BombRadius - TrueDistance ) / BombRadius;

          TileInfo& Struct = Structure( X, Y );

          if ( Struct.Strength == 0 )
          {
            DamageInnerStructure( X, Y, CurBombStrength );
          }
          else if ( CurBombStrength >= Struct.Strength )
          {
            // broken
            float   CurDamage = CurBombStrength;
            CurBombStrength -= Struct.Strength;
            DamageStructure( X, Y, CurDamage );
          }
          else
          {
            // apply damage
            DamageStructure( X, Y, CurBombStrength );
          }
        }
      }
    }
  }

}